{---------------------------------------------------------------------------}
{ BitMachine Program:  BITMACH.PAS                                          }
{ DATE:         (See Version, below)                                        }
{ PROGRAMMER:   Terry King                                                  }
{ WHAT IT DOES: Implements the BITMACHINE                                   }
{ HISTORY:                                                                  }
{ 01/14/97 - Update: set level to 1.00: Latest                              }
{ 06/20/98 - Quiet doesn't kill BEEP                                        }
{ 02/06/99 - Trying to find latest level!!                                  }
{ 02/21/99 - Adding Beep2, Adding QuietMODE save with .BIM                  }
{ 04/18/03 - Checking things out                                            }
{                                                                           }
{                                                                           }
{---------------------------------------------------------------------------}

{---------------------------------------------------------------------------}
{               U S E S :  External UNITS with needed code                  }
{---------------------------------------------------------------------------}

USES  dos, crt,   {Standard TURBO units}
      Parport,    {Parallel printer port I/O support}
      Mwindow,    {DOS Text windows}
      KbdInput,   {ReadLnPF support for text and keystrokes}
      HRTimers,   {High Resolution Timers}
      Sounds,     {Sound effects}
      Hexutil;    {Formatting utilities, Sounds}


{---------------------------------------------------------------------------}
{C1             C O N S T A N T S:                                          }
{---------------------------------------------------------------------------}
CONST    {NOTE: KeyStroke constants are in KBDINPUT.PAS}

Version     = '1.300 [04/18/03:   ]'; {++Program version}

MAXSTEP     = 100;  {Max number of steps in StepArray }
MAXIOBITS   = 8;    {Max number of bits in I/O Words  }
MAXCOUNTERS = 4;    {Max number of counters  }
MAXTIMERS   = 4;
MAXMSGS     = 20;   {Saved messages}

OUTPUTBYTESTART = $00; {Assigned to OutputByteInit}

      {Program Window location}
PWIN_L_ULX      =  2; {Large window}
PWIN_L_ULY      =  7;
PWIN_S_ULX      =  2; {Small window}
PWIN_S_ULY      = 17;
PWIN_LRX        = 79;
PWIN_LRY        = 23;

CHARCOLOR       = Yellow;   {Our screen colors}
BACKCOLOR       = Blue;
BLINKIT         = 128;
NOBLINK         = 0;
WAIT            = FALSE;  {For ReadLnPF}
NOWAIT          = TRUE;

APPEAR_NORM     = 1;   {Step line appearance: normal}
APPEAR_HIGH     = 2;   {HighLighted}
APPEAR_ERR      = 3;   {Error}

      {IO Name Window locations}
OUTIO_ULX       =  2; {OUTPUT Names}
OUTIO_ULY       =  6;
OUTIO_LRX       =  9;
OUTIO_LRY       = 13;

MSGFIELDSIZE    = 43;

INIO_ULX        = 71; {OUTPUT Names}
INIO_ULY        =  6;
INIO_LRX        = 77;
INIO_LRY        = 13;

{---------------------------------------------------------------------------}
{               T Y P E S :                                                 }
{----------------------------------------------------------------------------}

TYPE


FileNameTYPE    = String[50];  {File Name Strings}
MaxString       = String[100];
MsgString       = String[MSGFIELDSIZE];

STRING8TYPE      = string[8];
STRING32TYPE     = string[32];
STEPTYPE         = char;      {Type of StepString data}

STEPENTRY =
    RECORD
      StepName   : STRING8TYPE;  {Name of this step (may be null }
      NextStep   : STRING8TYPE;  {Name of this step (may be null }
      StepString : STRING32TYPE; {Step operands: Output, Input, commands }
      StepOper   : STEPTYPE;     {Type of StepString data: M, C, X}
    END;

IOENTRY =
    RECORD
      IOName     : STRING8TYPE;  {Name of I/O Bit}
      Inverted   : Boolean;      {Bit sense should be inverted }
    END;

BMSTATETYPE = (STOPPED, RUNNING, STEPPING, BREAKING, ENDED, ERRORSTOPPED);
{---------------------------------------------------------------------------}
{V1             V A R I A B L E S:                                          }
{---------------------------------------------------------------------------}
VAR {MainLine}
{-----------------( ARRAYS, COMPLEX DATA )-----------------}
StepArray      : ARRAY[1..MAXSTEP] OF STEPENTRY; {Array holding BitMachine Pgm}
StepSave       : STEPENTRY;                      {Var holding 1 entry}
InputNamesArray : ARRAY[0..(MAXIOBITS - 1)] OF IOEntry;
OutputNamesArray: ARRAY[0..(MAXIOBITS - 1)] OF IOEntry;

CounterArray    : ARRAY[1.. MAXCOUNTERS] OF LongInt;
TimerInUseArray : ARRAY[1.. MAXTIMERS] OF Boolean;

MsgSaveArray    : ARRAY[1.. MAXMSGS] OF String[MSGFIELDSIZE];

{---( HRTimerArray is in HRTimers UNIT )----}
CompareTime    : HRTime_rec_type;         {For comparison}
PgmDescription : String[50];              {Saved with file, shown at load time}

ResponseType   : ReadlnPF_TYPE;           {For use by KbdInput Unit }
KeyCode        : KeyCodeTYPE;
UserString     : CommandStringTYPE;
Statement      : CommandStringTYPE;

PgmFile        : Text;                    {Program load/save}
PgmFileName    : FileNameTYPE;
OldPgmFileName : FileNameTYPE;

UtilStr        : MaxString;
InWorkStr      : MaxString;               {Used in handling Input Data}
InDataStr      : MaxString;
OutWorkStr     : MaxString;               {Used in handling Output Data}
OldOutStr      : MaxString;
OutDataStr     : MaxString;

Msg            : MsgString;

KeyWord,
KeyValue,
KeyValue2,
KeyValue3,
Remainder      : MaxString;               {For #define, etc }

BMState        : BMSTATETYPE;             {Current state of the BitMachine}

{-----------------( SCALAR VARIABLES )----------------------------}
KbChar           : char;                    { for kbd input }
ChUtil           : char;
CounterStr       : String[1];               {COUNTERn }
TimerStr         : String[1];               {TIMERn }

XCur, YCur       : Byte;                    {User Cursor}

InputByte        : Byte;                    {Data read In/Out each SampleTime}
InWorkByte       : Byte;
OutputByte       : Byte;
OutWorkByte      : Byte;
InputByteInvert  : Byte;                    {Bits to optionally invert}
OutputByteInvert : Byte;
OutputByteInit   : Byte;                    {Startup time value}
Xnow, Ynow       : Byte;                    {Cursor location}
Bit2Edit         : Byte;


Step             : word;                    { Current Active BitMachine Step }
StepLoopBack     : word;                    { Step at beginning of this cycle}
StepJumpTo       : word;                    { Step to jump to via NextStep}
Step2Show        : Integer;                 { Step to show at top of window}
PgmTopStep       : word;                    { Top of program? }
LastEdit         : word;                    { Last edited line}

MsgWritePtr      : word;                    { Saved Msg}
MsgReadPtr       : word;
MsgMaxPtr        : word;                    { Highest Saved Msg }

ValueWord        : Word;                    {Command Value}
ValueWord2       : Word;
ValueWord3       : Word;
ValueLongInt     : longint;                 {For bigger values}

n                : word;                    {general purpose}
Appearance       : word;                    {Displayed line appearance}
Rtn              : word;                    {Return code from CALL}
RtnCode          : Integer;                 {General use}
SampleTime       : word;                    {Sample period (Ms)}

Samples          : Longint;                 {Samples of I/O }

ElapseduS,
CompareuS        : Real;


Done             : Boolean;                 {End main loop when Done}
SkipField        : Boolean;                 {Skip field being edited}
ShowNames        : Boolean;                 {Show the Input/Output Bit Names}
UpDateScreen     : Boolean;                 {Update user program area}
QuietMODE        : Boolean;                 {No Sounds }
EditCommand      : Boolean;                 {User entering a command, not data}
TestCommand      : Boolean;                 {User entering a test line}
IsTimerCmd       : Boolean;                 {Timer function is a command }
HaveChar         : Boolean;                 {Already have a Char in Ch}
TestLooping      : Boolean;                 {Test looping is underway}
WaitingForKey    : Boolean;                 {WAITKEY in progress}

Stop_Flag        : Boolean;                 {Error happened}
Debug            : Boolean;                 {Controls output}
TakeNextStep     : Boolean;                 { Take jump to NextStep field}
ForceNextStep    : Boolean;                 { PF Key forces step}

CounterCheck     : Boolean;                 {Counter comparison was true}
TimerCheck       : Boolean;                 {Timer comparison was true}
EditOut          : Boolean;                 {Output VS Input}


(*Ŀ
   INTERNAL PROCEDURES and FUNCTIONS                                       
  *)
{P1}


{------------------------------------------------------------------------}
{-----( ShowState PROCEDURE: )-------------------------------------------}
{------------------------------------------------------------------------}
PROCEDURE ShowState;

VAR Xnow, Ynow: Byte;
  BEGIN
    XNow := WhereX;
    YNow := WhereY;
    GoToXY(71,2);
    IF (BMState = RUNNING)      THEN Write('RUNNING ');
    IF (BMState = STOPPED)      THEN Write('STOPPED ');
    IF (BMState = STEPPING)     THEN Write('STEPPING');
    IF (BMState = BREAKING)     THEN Write('BREAKING');
    IF (BMState = ERRORSTOPPED) THEN Write('*ERROR* ');
    GoToXY(Xnow, Ynow);
  END;



{------------------------------------------------------------------------}
{-----( LookUpNextStep FUNCTION: Find StepNum of a Label         )-------}
{------------------------------------------------------------------------}
FUNCTION LookUpNextStep(Step2Find: STRING8TYPE): word;
VAR
  StepValue: word;
  BEGIN
    n         := 1;
    StepValue := 0;
    WHILE     (n <= PgmTopStep)
          AND (StepValue   = 0 ) DO
      BEGIN
        WITH StepArray[n] DO
            IF (Step2Find = StepName) THEN StepValue := n;
        INC(n);
      END;
    LookUpNextStep := StepValue;
  END;{LookUpNextStep}
{------------------------------------------------------------------------}
{-----( BlankMsg PROCEDURE: Clear bottom status area             )-------}
{------------------------------------------------------------------------}
PROCEDURE BlankMsg;

VAR Xnow, Ynow: Byte;
  BEGIN
    XNow := WhereX;
    YNow := WhereY;
    GoToXY(17,24);
    FOR N := 1 to MSGFIELDSIZE DO Write(' ');
    GoToXY(Xnow, Ynow);
  END;


{------------------------------------------------------------------------}
{-----( ShowMsg PROCEDURE: Write string to bottom status area    )-------}
{------------------------------------------------------------------------}
PROCEDURE ShowMsg(Msg:string);

VAR Xnow, Ynow: Byte;
  BEGIN
    XNow := WhereX;
    YNow := WhereY;
    GoToXY(17,24);
    BlankMsg;
    GoToXY(17,24);
    write(Msg);
    GoToXY(Xnow, Ynow);
  END;


{------------------------------------------------------------------------}
{-----( SaveMsg PROCEDURE: Write string to save array            )-------}
{------------------------------------------------------------------------}
PROCEDURE SaveMsg(Msg:string);

BEGIN
  INC(MsgWritePtr);
  IF (MsgWritePtr > MAXMSGS)
    THEN MsgWritePtr := 1
    ELSE MsgMaxPtr   := MsgWritePtr; {For retrieval}
  MsgSaveArray[MsgWritePtr] := Msg;
END;


{------------------------------------------------------------------------}
{-----( ByteOptProcess FUNCTION: Handle X, I, R  )----------------------}
{------------------------------------------------------------------------}
FUNCTION ByteOptProcess(DataStr: STRING8TYPE;
                     OldDataStr: STRING8TYPE): STRING8TYPE;

VAR

WorkStr   : STRING8TYPE;

  BEGIN
    WorkStr := '        ';
    FOR n := 1 TO MAXIOBITS DO
      BEGIN
        IF (DataStr[n] = '0' ) THEN WorkStr[n] := '0';
        IF (DataStr[n] = '1' ) THEN WorkStr[n] := '1';
        IF (DataStr[n] = 'X' ) THEN WorkStr[n] := OldDataStr[n];
        IF (DataStr[n] = 'I' )
          THEN
            BEGIN
              IF (OldDataStr[n] = '0' ) THEN WorkStr[n] := '1';
              IF (OldDataStr[n] = '1' ) THEN WorkStr[n] := '0';
            END;
        IF (DataStr[n] = 'R' )
          THEN
            BEGIN
              IF (Random(2) = 1)
                THEN WorkStr[n] := '1'
                ELSE WorkStr[n] := '0';
            END;
      END;

    ByteOptProcess := WorkStr;
  END;{ByteOptProcess}


{------------------------------------------------------------------------}
{-----( Comp8Bits FUNCTION: Compare 2 strings of 0,1,X             )----}
{------------------------------------------------------------------------}
FUNCTION Comp8Bits(DataStr: STRING8TYPE; CompStr: STRING8TYPE): Boolean;

VAR CompOK: Boolean;

  BEGIN
    n := 1;
    CompStr := Upper(CompStr); {Make sure uppercase X}
    CompOK := TRUE;
    WHILE CompOK AND (n <= MAXIOBITS) DO
      BEGIN
        IF     (DataStr[n] <> CompStr[n])
           AND (CompStr[n] <> 'X')
          THEN CompOK := FALSE;
          INC(n);
      END;

    Comp8Bits := CompOK;
  END;{Comp8Bits}

{------------------------------------------------------------------------}
{-----( ShowMsgOpt PROCEDURE: Write string to bottom status area    )----}
{------------------------------------------------------------------------}
PROCEDURE ShowMsgOpt(Msg:string; Color: byte; Attr: byte);

VAR Xnow, Ynow: Byte;
  BEGIN
    XNow := WhereX;
    YNow := WhereY;
    GoToXY(17,24);
    write('                                           ');
    GoToXY(17,24);
    TextColor(Color + Attr);
    write(Msg);
    TextColor(CHARCOLOR);
    GoToXY(Xnow, Ynow);
  END;

{----------------------------------------------------------------}
{--( InitArrays: Initialize: at startup, before load )-----------}
{----------------------------------------------------------------}
PROCEDURE InitArrays;

BEGIN
  {-----( Initialize StepArray )---}
  FOR Step := 1 TO MAXSTEP DO
    BEGIN
      WITH StepEntry(StepArray[Step]) DO
        BEGIN
          StepName   := '        ';
          NextStep   := '        ';
          StepString := 'XXXXXXXXXXXXXXXX                ';
          StepOper   := 'M';
        END;
    END; {Initialize StepArray}

  Step        := 1;
  Step2Show   := 1;
  PgmTopStep := 1;

  {-----( Initialize Input, Output Name Arrays )---}
  FOR n := 0 TO (MAXIOBITS - 1) DO
    BEGIN
      WITH InputNamesArray[n] DO
        BEGIN
          Str(n, UtilStr);
          IOName   := 'IN' + UtilStr + '     ';
          Inverted := FALSE;
        END;
    END;{Init Inputs}

  FOR n := 0 TO (MAXIOBITS - 1) DO
    BEGIN
      WITH OutputNamesArray[n] DO
        BEGIN
          Str(n, UtilStr);
          IOName   := 'OUT' + UtilStr + '    ';
          Inverted := FALSE;
        END;
    END;{Init Outputs}


  FOR n := 1 TO MAXCOUNTERS DO CounterArray[n] := 0;

  FOR n := 1 TO MAXTIMERS   DO TimerInUseArray[n] := FALSE;

END;{InitArrays)

{----------------------------------------------------------------}
{--( ToggleBit:  Toggle Output Bit )-----------------------------}
{----------------------------------------------------------------}
PROCEDURE ToggleBit(BitNum: word);

VAR BitMask :byte;

BEGIN
  BitMask := $01 SHL BitNum;
  IF ((BitMask AND OutputByte) <> 0)
    THEN OutputByte := OutputByte AND (NOT BitMask) {Turn bit off}
    ELSE OutputByte := OutputByte OR       BitMask; {Or on}
END;{ToggleBit}


{----------------------------------------------------------------}
{--( InitScreen: Write the initial screen )----------------------}
{----------------------------------------------------------------}
PROCEDURE InitScreen;
BEGIN
  TextBackground(BACKCOLOR);
  TextColor(CHARCOLOR);
  ClrScr;

  Writeln('Ŀ');
  Writeln('HelpLoadQuitEditRun StopStepBreak   Save                          ');
  Writeln(' F1  F2  F3  F4  F5  F6  F7  F8   F9 F10                          ');
  Writeln('Ĵ');
  Write  ('OUTPUT NAMES AND PIN NUMBERS  ');
  TextBackground(RED);
  TextColor(WHITE);
  Write  ('  The BitMachine!  ');
  TextBackground(BACKCOLOR);
  TextColor(CHARCOLOR);
  Writeln(' INPUT PIN NUMBERS AND NAMES');
  Writeln('Out0    (2)Ŀ                   (17)In7     ');
  Writeln('Out1    (3)Ŀ                     (16)In6     ');
  Writeln('Out2    (4)Ŀ                       (15)In5     ');
  Writeln('Out3    (5)Ŀ                         (14)In4     ');
  Writeln('Out4    (6)Ŀ                           (13)In3     ');
  Writeln('Out5    (7)Ŀ                             (12)In2     ');
  Writeln('Out6    (8)Ŀ                               (11)In1     ');
  Writeln('Out7    (9)Ŀ                                 (10)In0     ');
  Writeln('                                                             ');
  Writeln('StepName Step  OUTPUT                     INPUT      NextStep');
  Writeln('-------- ----                                                        --------');
  Writeln('                                                                             ');
  Writeln('                                                                             ');
  Writeln('                                                                             ');
  Writeln('                                                                             ');
  Writeln('                                                                             ');
  Writeln('                                                                             ');
  Writeln('Ĵ');
  Writeln('StepNow = 000                                             Samples =        ');
  Write  ('');

  GoToXY(2,17);
  ShowMsg(' BitMach   V: ' + Version );
END;{InitScreen}

{----------------------------------------------------------------}
{--( GetUserString  : Prompt, get string from user  )------------}
{----------------------------------------------------------------}
FUNCTION GetUserString(PromptStr : MaxString;
                       DefaultStr: MaxString;
                       Xloc      : Byte;
                       Yloc      : Byte)
                                 : MaxString;
VAR

Xnow, Ynow: Byte;
FieldSize : word;

BEGIN
  Xnow := WhereX;
  Ynow := WhereY;
  FieldSize := Length(PromptStr) + Length(DefaultStr) + 5;
  GoToXY(Xloc, Yloc);
  BlankMsg;
  GoToXY(Xloc, Yloc);
  Write(PromptStr,DefaultStr);
  ResponseType := ReadlnPF(UserString, KeyCode, FALSE );{Get user input}
  IF (ResponseType = Esc) THEN ;   {??}
  IF (ResponseType = CommandStr)
    THEN
      BEGIN
        IF (Length(UserString) > 1)
          THEN GetUserString := UserString
          ELSE GetUserString := DefaultStr;
      END;

  {--( Blank out where we were )----}
  GoToXY(Xloc, Yloc);
  BlankMsg;

  GoToXY(Xnow,Ynow);
END;{GetUserString}

{----------------------------------------------------------------}
{--( CleanLeadBlanks  : Remove leading blanks from a string )----}
{----------------------------------------------------------------}
PROCEDURE CleanLeadBlanks(VAR InStr: MaxString);

BEGIN
  WHILE (InStr[1] = ' ') AND (Length(InStr) > 0)
    DO Delete(InStr,1,1);
END;


{----------------------------------------------------------------}
{--( PadBlanksTo  : Pad with trailing blanks to desired length)--}
{----------------------------------------------------------------}
PROCEDURE PadBlanksTo(VAR InStr: MaxString; PadLength: word);

BEGIN
  WHILE (Length(InStr) < PadLength) DO InStr := Instr + ' ';
END;


{----------------------------------------------------------------}
{--( GetNext Procedure: Get next token )-------------------------}
{  InString : Input String with one or more tokens               }
{  Token    : Result String with first token found               }
{  RemainStr: String containing the rest of InString after Token }
{----------------------------------------------------------------}
Procedure GetNext(InString: maxstring;
                        ch: char;
                 Var token: maxstring;
            Var  RemainStr: maxstring);
Var
  SepLocation :integer;   {separator position}

BEGIN
  SepLocation := pos(ch,InString);                  {Locate 1st separator char}
  IF SepLocation <> 0 THEN                            {IF separator char found}
  BEGIN
    token := copy(InString,1,SepLocation);                      {Extract Token}
    delete(token,SepLocation,1);                   {Remove separator character}
    RemainStr := InString;
    delete(RemainStr,1,SepLocation);                 {Delete preseparator text}
    WHILE     (RemainStr[1]=ch)             {Remove additional separator chars}
          AND (length(RemainStr)>0)
      DO delete(RemainStr,1,1);
  END                                                       {A token was found}
  ELSE                                                {No separator char found}
  BEGIN
    token := InString;                           {Put entire string into token}
    RemainStr := '';                                        {RemainStr is null}
  END;
END;

{------------------------------------------------------------------}
{-----( ERROR PROCEDURE:  Print error and HALT )-------------------}
{------------------------------------------------------------------}
  procedure Error(msg: string);
    begin
      WriteLn('ERROR IN EXTERNAL PROGRAM: ' + msg);
      halt;

    end;

{------------------------------------------------------------------}
{    EXECANDCHECK PROCEDURE:  Execute an external program          }
{                                                                  }
{  NOTE: The MAIN program used MUST provide free memory for program}
{        to be execute by setting 'Maximum Heap Size' to a small   }
{        but useable value (say 10000)                             }
{                                                                  }
{------------------------------------------------------------------}

  procedure ExecAndCheck(Pgm: PathStr; CmdLine: string);
    begin
      SwapVectors;
      Exec(Pgm,CmdLine);
      SwapVectors;
      case DosError of
         2: Error('File not found: '+Pgm);
         8: Error('Insufficient memory to run '+Pgm);
        10: Error('Invalid environment');
        11: Error('Invalid format');
      end;
    end;


{------------------------------------------------------------------------}
{-----( CALL PROCEDURE: Execute external program by it's name only )-----}
{------------------------------------------------------------------------}

  procedure Call(PgmName: PathStr; CmdLine: ComStr; var ErrorLevel: word);

  VAR   PgmPath   : PATHSTR;
        PgmEnvPath : string;
    begin
      PgmEnvPath := GetEnv('PATH');
      PgmPath := FSearch(PgmName,PgmEnvPath);
      IF (Length(PgmPath) <> 0) THEN ExecAndCheck(PgmPath,CmdLine);
      ErrorLevel := DosExitCode;
(*      ShowMsg('Prog returned '+ Val(Errorlevel);  *)
    end;


{------------------------------------------------------------------------}
{-----( ShowCounter PROCEDURE: Show counter in center area  )------------}
{------------------------------------------------------------------------}
PROCEDURE ShowCounter(CounterNum: Word);

VAR Xnow, Ynow: Byte;
  BEGIN
    XNow := WhereX;
    YNow := WhereY;
    GoToXY(32,CounterNum + 5);
    write('COUNTER', CounterNum, ' ', CounterArray[CounterNum]:10);
    GoToXY(Xnow, Ynow);
  END;


{------------------------------------------------------------------------}
{-----( ShowTimer   PROCEDURE: Show Timer   in center area  )------------}
{------------------------------------------------------------------------}
PROCEDURE ShowTimer(TimerNum: HRTimer_TYPE);

VAR Xnow, Ynow: Byte;
    WorkStr   : String;
  BEGIN
    XNow := WhereX;
    YNow := WhereY;
    GoToXY(32,TimerNum + 9);
    ConvertHRTimers;
    WorkStr := HRTimeStr(TimerNum);
    write('TIMER', TimerNum, ' ', WorkStr );
    GoToXY(Xnow, Ynow);
  END;


{------------------------------------------------------------------------}
{-----( ClearCounterArea PROCEDURE: Clear the counter display  )---------}
{------------------------------------------------------------------------}
PROCEDURE ClearCounterArea;

VAR Xnow, Ynow: Byte;
  BEGIN
    XNow := WhereX;
    YNow := WhereY;
    FOR N := 1 to (MAXCOUNTERS+ MAXTIMERS) DO
      BEGIN
        GoToXY(32,n + 5);
        write('                   ');
      END;
    GoToXY(Xnow, Ynow);
  END;

{------------------------------------------------------------------------}
{-----( ShowSamples PROCEDURE: Write string to bottom status area    )---}
{------------------------------------------------------------------------}
PROCEDURE ShowSamples;

VAR Xnow, Ynow: Byte;
  BEGIN
    XNow := WhereX;
    YNow := WhereY;
    GoToXY(71,24);
    write(Samples:6);
    GoToXY(Xnow, Ynow);
  END;

(*
{------------------------------------------------------------------------}
{-----( ShowIONames  PROCEDURE: Write Names in Input/Output areas    )---}
{------------------------------------------------------------------------}
PROCEDURE ShowIONames;

VAR

Xnow, Ynow       : Byte;
InvertMask       : Integer;

    BEGIN
       Xnow := WhereX;
       YNow := WhereY;
    {--( Show OUTPUT Names )---}

       TextColor(WHITE);

       InvertMask := $01;
       FOR n := 0 to (MAXIOBITS - 1) DO
       BEGIN
         WITH OutputNamesArray[n] DO
           BEGIN
             GoToXY(OUTIO_ULX, (OUTIO_ULY + n) );
             write(IOName);
             IF Inverted
               THEN OutputByteInvert := OutputByteInvert OR Invertmask;
             InvertMask := InvertMask SHL 1;
           END;
       END;

    {--( Show INPUT Names )---}
       InvertMask := $01;

       FOR n := 0 to (MAXIOBITS - 1) DO
       BEGIN
         WITH InputNamesArray[n] DO
           BEGIN
             GoToXY(INIO_ULX, (INIO_ULY + (7-n)) );
             write(IOName);
             IF Inverted
               THEN InputByteInvert := InputByteInvert OR Invertmask;
             InvertMask := InvertMask SHL 1;
           END;
       END;
      TextBackground(BACKCOLOR);
      TextColor(CHARCOLOR);
      GoToXY(Xnow,Ynow);
    END;

*)


{------------------------------------------------------------------------}
{-----( ShowIONames  PROCEDURE: Write Names in Input/Output areas    )---}
{------------------------------------------------------------------------}
PROCEDURE ShowIONames;

VAR

Xnow, Ynow       : Byte;
InvertMask       : Integer;

    BEGIN
       Xnow := WhereX;
       YNow := WhereY;
    {--( Show OUTPUT Names )---}

       TextColor(LightCyan);

       InvertMask := $01;
       FOR n := 0 to (MAXIOBITS - 1) DO
       BEGIN
         WITH OutputNamesArray[n] DO
           BEGIN
             GoToXY(OUTIO_ULX, (OUTIO_ULY + n) );
             IF Inverted
               THEN
                 BEGIN
                   TextColor(LightMagenta);
                   write(IOName);
                   TextColor(LightCyan);
                   OutputByteInvert := OutputByteInvert OR Invertmask;
                 END
               ELSE
                 BEGIN
                   write(IOName);
                   OutputByteInvert := OutputByteInvert AND NOT Invertmask;
                 END;

             InvertMask := InvertMask SHL 1;
           END;
       END;

    {--( Show INPUT Names )---}
       InvertMask := $01;

       FOR n := 0 to (MAXIOBITS - 1) DO
       BEGIN
         WITH InputNamesArray[n] DO
           BEGIN
             GoToXY(INIO_ULX, (INIO_ULY + (7-n)) );
             IF Inverted
               THEN
                 BEGIN
                   TextColor(LightMagenta);
                   write(IOName);
                   TextColor(LightCyan);
                   InputByteInvert := InputByteInvert OR Invertmask;
                 END
               ELSE
                 BEGIN
                   write(IOName);
                   InputByteInvert := InputByteInvert AND NOT Invertmask;
                 END;

             InvertMask := InvertMask SHL 1;
           END;
       END;
      TextBackground(BACKCOLOR);
      TextColor(CHARCOLOR);
      GoToXY(Xnow,Ynow);
    END;


{------------------------------------------------------------------------}
{-----( ShowStepNum PROCEDURE: Write step number in lower area    )------}
{------------------------------------------------------------------------}
PROCEDURE ShowStepNum;

VAR ValStr: String[4];
  BEGIN
    XNow := WhereX;
    YNow := WhereY;
    GoToXY(11,24);
    Str(Step:4, ValStr);
    TextColor(LightRed);
    write(ValStr);
    TextColor(Yellow);
    GoToXY(Xnow, Ynow);
  END;

{------------------------------------------------------------------------}
{-----( ExecuteIO   PROCEDURE: Do Digital Input/Output  )--------------0--}
{------------------------------------------------------------------------}
PROCEDURE ExecuteIO;

  BEGIN
    WriteDO(OutputByte XOR OutputByteInvert);
    InputByte := ReadDI XOR InputByteInvert;
  END;


{------------------------------------------------------------------------}
{-----( ExecuteStep PROCEDURE: Do the current Step(s)   )--------------0--}
{------------------------------------------------------------------------}
PROCEDURE ExecuteStep;

  BEGIN

    WITH StepArray[Step] DO
      BEGIN
        CASE StepOper OF
          'M':   {Machine Step: Operate on Output Bit String}
            BEGIN
              {---( Convert Output Data string to OutputByte )-------}
              OutWorkStr  := Copy(StepString,1,8);
              OldOutStr   := Byte2Bin(OutPutByte);            {Get current bits}
              OutDataStr  := ByteOptProcess(OutWorkStr, OldOutStr); {Do X,I,R}
              OutPutByte  := Bin2Byte(OutDataStr, RtnCode);
              {---( Check if we might loop back here )---------------}
              InWorkStr    := Copy(StepString,9,8);
              IF (InWorkStr <> 'XXXXXXXX')
                THEN StepLoopBack := Step
                ELSE StepLoopBack := 0;
            END;{'M'}

          'T':  {--------( Test-Only Line:            )-----------}
            BEGIN
              IF (StepLoopBack = 0) THEN StepLoopBack := Step;{Start of a loop}
            END;{'T'}


          'C':  {--------( Internal Command: Execute it )-----------}
            BEGIN
{??}         TakeNextStep := TRUE;
             GetNext(StepString, ' ', KeyWord, Remainder);
             GetNext(Remainder , ' ', KeyValue, Remainder);
(*
             ShowMsg('C:: KEY=' + Keyword + '  VAL=' + KeyValue);
*)
     {--( WAIT )-----}
             IF (KeyWord = 'WAIT')
               THEN
                 BEGIN
                   Val(KeyValue, ValueWord, RtnCode);
                   Delay(ValueWord);
                 END;

     {--( WAITRANDOM )-----}
             IF (KeyWord = 'WAITRANDOM')
               THEN
                 BEGIN
                   Randomize;
                   Val(KeyValue, ValueWord, RtnCode);
                   Delay(Random(ValueWord));
                 END;

     {--( WAITKEY )-----}
             IF (KeyWord = 'WAITKEY')
               THEN
                 BEGIN
                   IF WaitingForKey
                   THEN
                     BEGIN
                       IF KeyPressed
                       THEN
                         BEGIN
                           WaitingForKey := FALSE;
                           BlankMsg;
                         END
                       ELSE DEC(Step);
                     END
                   ELSE
                     BEGIN
                       ShowMsg('** HIT ANY KEY TO CONTINUE **');
                       WaitingForKey := TRUE;
                     END;
                 END;

     {--( BEEP )-----}
             IF (KeyWord = 'BEEP')
               THEN
                 BEGIN
                   IF (KeyValue = '')  THEN BeepHi;
                   IF (KeyValue = '1') THEN BeepLo;
                   IF (KeyValue = '2') THEN PassTone;
                   IF (KeyValue = '3') THEN FailTone;
                   IF (KeyValue = '4') THEN IntErrorTone;
                   IF (KeyValue = '5') THEN Click;
                   IF (KeyValue = '6') THEN Beep2;
                 END;

     {--( SOUND )-----}
             IF (KeyWord = 'SOUND')
               THEN
                 BEGIN
                   GetNext(Remainder , ' ', KeyValue2, Remainder);
                   Val(KeyValue2, ValueWord2, RtnCode);
                   IF (RtnCode <> 0) THEN ValueWord2 := 1;

                   {--------( Handle individual Sounds )---------}

                   IF (KeyValue = 'ATTNTONE') THEN AttnTone;

                   IF (KeyValue = 'INTERRORTONE') THEN IntErrorTone;

                   IF (KeyValue = 'FAILTONE') THEN FailTone;

                   IF (KeyValue = 'PASSTONE') THEN PassTone;

                   IF (KeyValue = 'PHONERING') THEN
                     BEGIN
                       PhoneRing(ValueWord2);
                     END;

                   IF (KeyValue = 'RINGING') THEN
                     BEGIN
                       Ringing(ValueWord2);
                     END;

                   IF (KeyValue = 'BUSY') THEN
                     BEGIN
                       Busy(ValueWord2);
                     END;

                   IF (KeyValue = 'WASHING') THEN
                     BEGIN
                       washing(ValueWord2);
                     END;

                   IF (KeyValue = 'DRIPS') THEN
                     BEGIN
                       drips(ValueWord2);
                     END;

                   IF (KeyValue = 'FILLING') THEN
                     BEGIN
                       filling(ValueWord2);
                     END;


                   IF (KeyValue = 'SPINNING') THEN
                     BEGIN
                       Spinning(ValueWord2);
                     END;

                 END;


     {--( COUNTER )--}
             IF (Copy(KeyWord,1,7) = 'COUNTER')
               THEN
                 BEGIN
                   CounterCheck := FALSE;
{??}               TakeNextStep := TRUE;
                   CounterStr := Copy(KeyWord,8,1);
                   Val(CounterStr, ValueWord, RtnCode);
                   IF    (RtnCode  <> 0)
                      OR (ValueWord < 0)
                      OR (ValueWord > MAXCOUNTERS)
                   THEN
                     BEGIN
                       Msg :='COUNTER:: NUMBER BAD!! = ' + Keyword;
                       ShowMsg(Msg);
                       SaveMsg(Msg);
                       BMState := ERRORSTOPPED;
                       RtnCode := -1;
                     END;

                   IF (RtnCode = 0) THEN
                     BEGIN
                       GetNext(Remainder , ' ', KeyValue2, Remainder);
(*
                       ShowMsg(Keyword + ' ' + KeyValue + ' ' + KeyValue2);
*)
                    {-----------------------------}
                       IF (KeyValue = 'RESET')
                       THEN
                         BEGIN
                           CounterArray[ValueWord] := 0;
                           ShowCounter(ValueWord);
                         END
                       ELSE
                    {-----------------------------}
                         BEGIN        {must be a +-<>=}
                           IF     (Length(KeyValue) = 1) THEN
                             ChUtil := KeyValue[1];
                           IF ChUtil IN ['+','-','>','<','='] THEN
                             BEGIN
                               TakeNextStep := FALSE;
                               Val(KeyValue2, ValueLongInt, RtnCode);
                               IF (RtnCode <> 0) THEN
                                 BEGIN
                                   Msg := 'COUNTER:: PARAMETER BAD!! = ' +
                                            KeyValue2;
                                   ShowMsg(Msg);
                                   SaveMsg(Msg);
                                 END;

                               IF (KeyValue = '+') THEN
                                 BEGIN
                                   INC((CounterArray[ValueWord]), ValueLongInt);
                                   ShowCounter(ValueWord);
                                   CounterCheck := TRUE;
                                 END;
                               IF (KeyValue = '-') THEN
                                 BEGIN
                                   DEC((CounterArray[ValueWord]), ValueLongInt);
                                   ShowCounter(ValueWord);
                                   CounterCheck := TRUE;
                                 END;
                               IF (KeyValue = '=') THEN
                                 BEGIN
                                   IF(CounterArray[ValueWord] = ValueLongInt)
                                     THEN CounterCheck := TRUE;
                                 END;
                               IF (KeyValue = '>') THEN
                                 BEGIN
                                   IF(CounterArray[ValueWord] > ValueLongInt)
                                     THEN CounterCheck := TRUE;
                                 END;
                               IF (KeyValue = '<') THEN
                                 BEGIN
                                   IF(CounterArray[ValueWord] < ValueLongInt)
                                     THEN CounterCheck := TRUE;
                                 END;
                             {----( Take Jump or continue. Don't hold )---}
                               IF CounterCheck
                               THEN TakeNextStep := TRUE
                               ELSE INC(Step);

                             END;{+-<>=}
                         END;{RESET OR +-<>=}

                     END;{Valid Counter Value}
                 END;{Counter}


     {--( TIMER )--}
             IF (Copy(KeyWord,1,5) = 'TIMER')
               THEN
                 BEGIN
                   IsTimerCmd := FALSE;
                   TimerCheck := FALSE;
{??}               TakeNextStep := TRUE;
                   TimerStr := Copy(KeyWord,6,1);
                   Val(TimerStr, ValueWord, RtnCode);
                   IF    (RtnCode  <> 0)
                      OR (ValueWord < 0)
                      OR (ValueWord > MAXTIMERS)
                   THEN
                     BEGIN
                       Msg := 'TIMER:: NUMBER BAD!! = ' + Keyword;
                       ShowMsg(Msg);
                       SaveMsg(Msg);
                       BMState := ERRORSTOPPED;
                       RtnCode := -1;
                     END;

                   IF (RtnCode = 0) THEN
                     BEGIN
                       TimerInUseArray[ValueWord] := TRUE;
                       GetNext(Remainder , ' ', KeyValue2, Remainder);
(*
                       ShowMsg(Keyword + ' ' + KeyValue + ' ' + KeyValue2);
*)
                 {----------------------------}
                       IF (KeyValue = 'RESET')
                       THEN
                         BEGIN
                           IsTimerCmd := TRUE;
                           StartHRTimer(ValueWord);
                           StrobeHRTimers;
                           ConvertHRTimers;
                           ShowTimer(ValueWord);
                         END;

                 {----------------------------}
                       IF (KeyValue = 'START')
                       THEN
                         BEGIN
                           IsTimerCmd := TRUE;
                           ReStartHRTimer(ValueWord);
                           StrobeHRTimers;
                           ConvertHrTimers;
                           ShowTimer(ValueWord);
                         END;

                 {----------------------------}
                       IF (KeyValue = 'STOP')
                       THEN
                         BEGIN
                           IsTimerCmd := TRUE;
                           StrobeHRTimers;
                           ConvertHRTimers;
                           StopHRTimer(ValueWord);
                           ShowTimer(ValueWord);
                         END;
                 {----------------------------}
                       IF NOT IsTimerCmd THEN
                         BEGIN        {must be a <> }
                           IF     (Length(KeyValue) = 1) THEN
                             ChUtil := KeyValue[1];
                           IF ChUtil IN ['>','<'] THEN
         {??}                BEGIN
                               TakeNextStep := FALSE;
                               ElapseduS := GetHRElapsedTimeUs(ValueWord);
                               CompareuS := HMSStr2Real(KeyValue2);
                               IF (KeyValue = '>') THEN
                                 BEGIN
                                   IF (ElapseduS > CompareuS)
                                     THEN TakeNextStep := TRUE;
                                 END;
                               IF (KeyValue = '<') THEN
                                 BEGIN
                                   IF (ElapseduS < CompareuS)
                                     THEN TakeNextStep := TRUE;
                                 END;
                             {----( Take Jump or continue. Don't hold )---}
                               IF TimerCheck
                               THEN TakeNextStep := TRUE
                               ELSE INC(Step);

                             END;{<>}
                         END;{RESET,STOP, START OR <>}

                     END;{Valid Timer Value}
                 END;{Timer}

     {--( STOP )-----}
             IF (KeyWord = 'STOP') THEN BMState := STOPPED;

     {--( STEP )-----}
             IF (KeyWord = 'STEP')
               THEN
                 BEGIN
                   IF (KeyValue = 'ON')  THEN BMState := STEPPING;
                   IF (KeyValue = 'OFF') THEN BMState := RUNNING;
                 END;


     {--( SHOW )-----}
             IF (KeyWord = 'SHOW')
               THEN
                 BEGIN
                   ShowMsg(KeyValue + ' ' + Remainder);
                 END;

            END;{'C'}

          'E':  {------( External Command: Shell to DOS and run it )-------}
            BEGIN
              GetNext(StepString, ' ', KeyWord, Remainder);
              ShowMsg('E:: KEY=' + Keyword + '  VAL=' + Remainder);
              Call(KeyWord, Remainder, Rtn);
              TakeNextStep := TRUE;
            END;{'E'}

        END;{CASE StepOper}

            {---( TEST: TAKE NEXT STEP ????           )------}
            {---( Input Data string VS Compare String )------}
            IF    (StepOper = 'M')
               OR (StepOper = 'T')
              THEN
                BEGIN
                  InWorkStr    := Copy(StepString,9,8);
                  InDataStr    := Byte2Bin(InputByte);
                  TakeNextStep := Comp8Bits(InDataStr, InWorkStr);
                  IF ForceNextStep THEN  {User hit <Ctrl_F7}
                    BEGIN
                      TakeNextStep  := TRUE;
                      ForceNextStep := FALSE;
                    END;
                END;

(*??          ELSE TakeNextStep := TRUE; {Must be an Int or Ext Command } *)

      {----( TAKE the next step, or NOT ??? )----------------------------}
            IF TakeNextStep
              THEN
               IF (NextStep = '        ') OR
                  (NextStep = '+       ')
                 THEN INC(Step)
                 ELSE
                   BEGIN
                     StepJumpTo := LookUpNextStep(NextStep);
                     IF (StepJumpTo <> 0)
                       THEN
                         Step := StepJumpTo
                       ELSE
                         BEGIN
                           BMState := ERRORSTOPPED;
                           UtilStr := 'NEXTSTEP "' + NextStep +
                                      '" NOT FOUND!!';
                           ShowMsgOpt(UtilStr, LightRed, BLINKIT);
                           SaveMsg(UtilStr);
                         END;
                   END{LookUpNextStep}
     {---( Else Test-Looping so take next step or loop back. )----}
              ELSE
                IF (StepOper = 'M') OR (StepOper = 'T')
                THEN IF (StepArray[Step + 1].StepOper = 'T') {Another test}
                     THEN INC(Step)
                     ELSE Step := StepLoopBack;             {End of loop: go back}



      END;{WITH}

      Step2Show := Step - 2;
      IF (Step2Show < 1) THEN Step2Show := 1;   {??}

      UpdateScreen := TRUE;

      IF (BMState = STEPPING)     THEN BMState := STOPPED;
      IF (Step > PgmTopStep)     THEN BMState := ENDED;

  END;{ExecuteStep}

{------------------------------------------------------------------------}
{-----( Show8Bits   PROCEDURE: Display 'LED' data      )-----------------}
{------------------------------------------------------------------------}
PROCEDURE Show8Bits(Data: Byte);

VAR

OutString: MAXSTRING;
n        : word;

BEGIN
  OutString := Byte2Bin(Data);
  TextBackGround(BACKCOLOR);
  FOR n := 1 to 8 DO
    BEGIN
      IF (OutString[n] = '1')
      THEN
        BEGIN
          TextColor(LightGreen);
          Write(CHR(219));
        END

      ELSE
        BEGIN
          TextColor(DarkGray);
          Write(Chr(219));
        END;
      TextColor(CHARCOLOR);
        Write(' ');
    END;
  TextBackGround(BACKCOLOR);
  TextColor(CHARCOLOR);
END;

{------------------------------------------------------------------------}
{-----( AddStepEntry: Put a StepEntry in StepArray  )--------------------}
{------------------------------------------------------------------------}

PROCEDURE AddStepEntry(StepPos : word;
                       Name    : STRING8TYPE;
                       Next    : STRING8TYPE;
                       StepStr : STRING32TYPE;
                       SType   : STEPTYPE);

BEGIN
  WITH StepEntry(StepArray[StepPos]) DO
    BEGIN
      StepName   := Name;
      NextStep   := Next;
      StepString := StepStr;
      StepOper   := SType;
    END;
END;

{------------------------------------------------------------------------}
{-----( InsertStepEntry: Insert a StepEntry into StepArray  )------------}
{------------------------------------------------------------------------}
PROCEDURE InsertStepEntry(StepPos : word);
BEGIN
  FOR n := PgmTopStep DOWNTO StepPos DO   {--( Move the program up 1 pos )--}
    BEGIN
      StepArray[n +1] := StepArray[n]
    END;

  WITH StepEntry(StepArray[StepPos]) DO   {--( Insert a 'normal blank' line)}
    BEGIN
      StepName   := '        ';
      NextStep   := '        ';
      StepString := 'XXXXXXXXXXXXXXXX                ';
      StepOper   := 'M';
    END;
  INC(PgmTopStep);
END;

{------------------------------------------------------------------------}
{-----( DeleteStepEntry: Delete a StepEntry from StepArray  )------------}
{------------------------------------------------------------------------}
PROCEDURE DeleteStepEntry(StepPos : word);
BEGIN
  FOR n := StepPos TO PgmTopStep DO   {--( Move the program down 1 pos )--}
    BEGIN
      StepArray[n] := StepArray[n + 1]
    END;
  DEC(PgmTopStep);
END;

{------------------------------------------------------------------------}
{-----( CopyStepEntry: Copy a StepEntry from StepArray to StepSave  )----}
{------------------------------------------------------------------------}
PROCEDURE CopyStepEntry(StepPos : word);
BEGIN
  StepSave := StepArray[StepPos];
END;

{------------------------------------------------------------------------}
{-----( PasteStepEntry: Paste a StepEntry from StepSave to StepArray )---}
{------------------------------------------------------------------------}
PROCEDURE PasteStepEntry(StepPos : word);
BEGIN
  StepArray[StepPos] := StepSave;
END;

{------------------------------------------------------------------------}
{-----( ShowStepEntry: Write a StepEntry on the display)-----------------}
{------------------------------------------------------------------------}

PROCEDURE ShowStepEntry(StepPos : word; Appearance: word);

VAR ValStr: String[4];
    N     : word;

BEGIN
  WITH StepEntry(StepArray[StepPos]) DO
    BEGIN
      IF (Appearance = APPEAR_HIGH ) THEN TextBackGround(LightCyan);
      IF (Appearance = APPEAR_ERR  ) THEN TextBackGround(RED);
      Str(StepPos:4, ValStr);

      Write(StepName + ' ' + ValStr + '  ');     {Stepname and Step}

      IF (StepOper = 'M') THEN                   {Input / Output}
        BEGIN
          FOR n := 1 to 8 DO
            BEGIN
              Write(StepString[n], ' ');
            END;
          Write('                  ');
          FOR n := 1 to 8 DO
            BEGIN
              Write(StepString[n+8], ' ');
            END;
        END;


      IF (StepOper = 'T') THEN                   {TEST ONLY}
        BEGIN
          Write('TEST ---------->                  ');      {Internal}
          FOR n := 1 to 8 DO
            BEGIN
              Write(StepString[n+8], ' ');
            END;
        END;



      IF (StepOper = 'C') OR (StepOper = 'E') THEN    {Command}
        BEGIN
          IF (StepOper = 'C') THEN Write('C: ');      {Internal}
          IF (StepOper = 'E') THEN Write('E: ');      {External}

          write(StepString);
          write('               ');
        END;

      Write('    ', NextStep);                          {NextStep}
    END;
  TextBackground(BACKCOLOR);
END;

{------------------------------------------------------------------------}
{-----( EditStepEntry: Edit a StepEntry on the display)------------------}
{------------------------------------------------------------------------}

PROCEDURE EditStepEntry;

VAR
  WorkStep   : STEPENTRY;
  ValStr     : String[4];
  WorkStr    : String[32];
  N          : word;
  Done       : Boolean;
  Ch         : Char;

BEGIN
   Done := FALSE;
   UserString := '';
   WorkStep := StepArray[Step2Show+2];
   LastEdit := Step2Show + 2;
   IF ((Step2Show + 2) > PgmTopStep) THEN PgmTopStep := Step2Show +2;
   (*??*) Step := PgmTopStep; {DEBUG}
   Window(PWIN_S_ULX, (PWIN_S_ULY + 2), PWIN_LRX, (PWIN_S_ULY + 2));

   WITH WorkStep DO
    BEGIN
    {--( EDIT StepName )--}
      GoToXY(1,1);
      SkipField := FALSE;
      ResponseType := ReadlnPF(UserString, KeyCode, FALSE );{Get user input}
      IF (ResponseType = Esc) THEN Done := TRUE;
      IF NOT Done AND (ResponseType <> Tab) THEN
        BEGIN
          IF (ResponseType = CommandStr) THEN
            UserString := Upper(UserString);
            IF (Length(UserString) > 0) THEN   {No action if just 'enter'}
              BEGIN
                WHILE (Length(UserString) < 8)
                  DO UserString := UserString + ' ';
                StepName := UserString;
                UserString := '';
              END;
        END; {Not Done}

    {--( CHECK IF OUTPUT OR INTERNAL/EXTERNAL COMMAND )--}
(*      StepOper := 'M';  {Default}
*)
      WorkStr  := '';
      Ch       := ' ';
      GoToXY(16,1);
      Ch        := ReadKey;  {First Char}

      IF     (Ch = Chr(9) )   {Tab}
         OR  (Ch = Chr(13))   {Enter}
        THEN SkipField := TRUE
        ELSE SkipField := FALSE;

      IF     (Ch = 'C')
         OR  (Ch = 'c')
         OR  (Ch = 'E')
         OR  (Ch = 'e')
      THEN EditCommand := TRUE ELSE EditCommand := FALSE;

      IF     (Ch = 'T')
         OR  (Ch = 't')
      THEN TestCommand := TRUE ELSE TestCommand := FALSE;

      {--------( EDIT OUTPUT )---------}
      IF    (NOT EditCommand)
        AND (NOT SkipField)
        AND (NOT TestCommand) THEN
        BEGIN
          StepOper := 'M';
          HaveChar := TRUE;
          REPEAT
            IF NOT HaveChar THEN Ch := ReadKey;
            HaveChar := FALSE;
            IF (Ch <> chr(9)) and (Ch <> chr(13)) THEN
              BEGIN
                Ch := UpCase(Ch);
                WorkStr   := WorkStr + Ch;
                Write(Ch + ' ');
              END;
          UNTIL     (Ch = chr(9))
                 OR (Ch = chr(13))           {Enter or Tab}
                 OR (Length(WorkStr) = 8);

          IF (Length(WorkStr) > 0) THEN
            WHILE (Length(Workstr) < 8)      {Fill out the string}
              DO
                BEGIN
                  IF (WorkStr[1] = ' ')
                  THEN WorkStr := WorkStr + ' '
                  ELSE WorkStr := WorkStr + 'X';
                END
              ELSE WorkStr := Copy(StepString,1,8);  {Don't change it}
        END;{Edit OUTPUT}

      {--------( TEST COMMAND )---------}
      IF TestCommand  THEN
        BEGIN
          StepOper := 'T';     {--( IS A TEST COMMAND )--}
          WorkStr := 'TEST -->';
          Write(     'TEST ---------->');
        END;

      {--------( EDIT INPUT )---------}
      IF    (NOT EditCommand)
        AND (NOT SkipField)
        OR  (TestCommand)
      THEN
        BEGIN
        {--( EDIT INPUT )--}
          GoToXY(50,1);
            REPEAT
              Ch        := ReadKey;
              IF (Ch <> chr(9)) and (Ch <> chr(13)) THEN
                BEGIN
                  Ch := UpCase(Ch);
                  WorkStr   := WorkStr + Ch;
                  Write(Ch + ' ');
                END;
            UNTIL     (Ch = chr(9))
                   OR (Ch = chr(13))           {Enter or Tab}
                   OR (Length(WorkStr) = 16);

            IF (Length(WorkStr) >= 8) THEN
              WHILE (Length(Workstr) <16 )     {Fill out the string}
                DO WorkStr := WorkStr + 'X'
                ELSE WorkStr := WorkStr + Copy(StepString,9,8);  {Don't change it}

            WHILE (Length(WorkStr) < 32) DO WorkStr := WorkStr + ' ';{Fill}
        END;{EDIT INPUT}


      {---( Done building WorkString: Put it in array element )---}
        IF (Length(WorkStr) > 0) THEN StepString := WorkStr;

      {--------( EDIT COMMANDS )---------}
      IF EditCommand THEN
        BEGIN

          IF     (Ch = 'C')
             OR  (Ch = 'c')
          THEN
            BEGIN
              StepOper := 'C';     {--( IS AN INTERNAL COMMAND )--}
              Write('C: ');
            END;

          IF     (Ch = 'E')
             OR  (Ch = 'e')
          THEN
            BEGIN
              StepOper := 'E';     {--( IS AN EXTERNAL COMMAND )--}
              Write('E: ');
            END;

          WorkStr := '';
          REPEAT
              BEGIN
                Ch := ReadKey;
                WorkStr   := WorkStr + Ch;
                Write(Ch );
              END;
          UNTIL     (Ch = chr(9))
                 OR (Ch = chr(13))           {Enter or Tab}
                 OR (Length(WorkStr) = 32);
          Delete(WorkStr,Length(WorkStr),1);            {Delete last char}
          WHILE (Length(WorkStr) < 32) DO WorkStr := WorkStr + ' '; {Fill}
          write('                 ');       {Space over to NextStep Field}

          StepString := Upper(WorkStr);            {Put in Array}

        END; {EditCommand}

    {--( EDIT NextStep )--}
      GoToXY(70,1);
      ResponseType := ReadlnPF(UserString, KeyCode, FALSE );{Get user input}
      IF (ResponseType = Esc) THEN Done := TRUE;
      IF NOT Done AND (ResponseType <> Tab) THEN
        BEGIN
          IF (ResponseType = CommandStr) THEN
            IF (Length(UserString) > 0) THEN   {No action if just 'enter'}
              BEGIN
                WHILE (Length(UserString) < 8)
                  DO UserString := UserString + ' ';
                NextStep := Upper(UserString);
                UserString := '';
              END;
        END; {Not Done}

   {---( Update data in Program Array, update the screen for use )---}
      StepArray[Step2Show+2] := WorkStep;
      UpdateScreen := TRUE;
    END;{WITH}

  Window(1,1,80,25);

END;{EditStepentry}


{------------------------------------------------------------------------}
{--( LoadPgm    PROCEDURE: Load a BitMachine Program from File   )-------}
{------------------------------------------------------------------------}
Procedure LoadPgm(FileName: FileNameTYPE);

VAR
StepString  : MaxString;
StepNameStr : MaxString;
CommandStr  : MaxString;
NextStepStr : MaxString;
OperTypeStr : Char;


Token       : MaxString;
Remainder   : MaxString;

NumPgmLines    : word;    {Actual program lines}
NumLines       : word;    {Total ALL lines}

IOBitNum       : word;    { Extracted IO Bit Number}

IsDefineLine   : Boolean;

BEGIN
  Stop_Flag := FALSE;
  Debug := FALSE;
  Assign(PgmFile,Filename);
  {$I-} Reset(PgmFile) {$I+};
  IF (IOresult<>0) THEN
    BEGIN
      Msg := '**ERROR**  Cannot find file ' + Filename;
      ShowMsgOpt(Msg,LightRed,0);
      SaveMsg(Msg);
      Stop_Flag := true;
    END
  ELSE {It's Cool: Go ahead}
    IF Debug THEN WriteLn;
    NumPgmLines := 0;
    IF NOT Stop_Flag THEN
      BEGIN
        WHILE (NOT Eof(PgmFile)) AND (NOT Stop_Flag)  DO
        BEGIN {Processing File Lines}

          Readln(PgmFile,StepString);      {Read TC statement}
          NumLines := NumLines + 1;

          {------( Check for #define line )---------}
          IF ( Copy(StepString, 1, 8) = '#define ' )
          THEN
            BEGIN
              IsDefineLine := TRUE;
              KeyWord  := '';
              KeyValue := '';
              {---( Delete #define )----}
              Delete(StepString,1,8);

              CleanLeadBlanks(StepString);
              GetNext(StepString, ' ', KeyWord, Remainder);

              {---------( Have a KeyWord. What to do?? )--------}
         (*     ShowMsg(KeyWord);*)

     {----( PgmDescription Keyword )--------------------}
              IF (KeyWord = 'PgmDesc') THEN
                BEGIN
                  PgmDescription := Remainder;
                  Msg := 'PgmDesc: ' + PgmDescription;
                  ShowMsg(Msg);
                  SaveMsg(Msg);
                END;

     {----( QuietMODE  Keyword )--------------------}
              IF (KeyWord = 'QuietMODE') THEN
                BEGIN
                  IF (Remainder = 'TRUE') THEN QuietMODE := TRUE;
                  IF (Remainder = 'FALSE') THEN QuietMODE := FALSE;
                END;

     {----( LastEdit Keyword )--------------------}
              IF (KeyWord = 'LastEdit') THEN
                BEGIN
                  GetNext(Remainder, ' ', KeyValue, Remainder);
                  Val(KeyValue,LastEdit, RtnCode);
                  Step2Show := LastEdit - 2;
                END;

     {----( SampleTime Keyword )--------------------}
              IF (KeyWord = 'SampleTime') THEN
                BEGIN
                  GetNext(Remainder, ' ', KeyValue, Remainder);
                  Val(KeyValue,SampleTime, RtnCode);
                END;

     {----( Handle 'OUT' Keyword )--------------------}
              IF ((Copy(KeyWord,1,3)) = 'OUT') THEN
                BEGIN
                  Val(Copy(StepString,4,1),IOBitNum, RtnCode);

                  PadBlanksTo(Remainder, 8);
                  WITH OutputNamesArray[IOBitNum] DO
                    BEGIN
                      IOName   := Remainder;
                      IF    (Length(StepString) = 15)
                        AND (StepString[15] = 'I')
                        THEN Inverted := TRUE
                        ELSE Inverted := FALSE;
                    END;
                END;

     {----( Handle 'IN' Keyword )--------------------}
              IF ((Copy(KeyWord,1,2)) = 'IN') THEN
                BEGIN
                  Val(Copy(StepString,3,1),IOBitNum, RtnCode);

                  PadBlanksTo(Remainder, 8);
                  WITH InputNamesArray[IOBitNum] DO
                    BEGIN
                      IOName   := Remainder;
                      IF    (Length(StepString) = 14)
                        AND (StepString[14] = 'I')

                        THEN Inverted := TRUE
                        ELSE Inverted := FALSE;
                    END;
                END;

      {--------------}
            END{Handle #define}
          ELSE
            BEGIN
              IsDefineLine := FALSE;
            END;
{---------------( Load program steps )--------------------------}

          IF NOT IsDefineLine THEN
            BEGIN   {------( Handle A Normal Line )-----------}
              IF   ( Length(StepString) = 32+8+8+1)
              THEN
                BEGIN
                  NumPgmLines := NumPgmLines + 1;
                {-------( Extract Strings from StepString )---}
                  StepNameStr    := Copy(StepString, 1, 8);
                  NextStepStr    := Copy(StepString, 9, 8);
                  CommandStr     := Copy(StepString, 17, 32);
                  OperTypeStr    := StepString[49];

                  IF Debug THEN
                    WriteLn(StepnameStr+'|'+CommandStr+'|'+
                            NextStepStr+'|'+OperTypeStr);

                  AddStepEntry(NumPgmLines,
                               StepNameStr,
                               NextStepStr,
                               CommandStr,
                               OperTypeStr);

                END
              ELSE
                BEGIN
                  Writeln('!!ERROR!!  File Format Bad:=>', StepString);
                END;
            END;

          ClearCounterArea;
          UpDateScreen := TRUE;
          PgmTopStep  := NumPgmLines;
          Step         := 2; {??}
{---------------( END: Load program steps )--------------------------}


        END;{Processing Load File Lines}
      IF NOT Stop_Flag THEN Close(PgmFile);
     END;{NOT Stop_Flag}

END;{LoadPgm}



{------------------------------------------------------------------------}
{--( SavePgm    PROCEDURE: Save a BitMachine Program to   File   )-------}
{------------------------------------------------------------------------}
Procedure SavePgm(FileName: FileNameTYPE);

VAR
PgmLine     : MaxString;
StepNameStr : MaxString;
CommandStr  : MaxString;
NextStepStr : MaxString;
OperTypeStr : MaxString;

NumLines    : word;

BEGIN
  Debug := FALSE;
  Assign(PgmFile,Filename);
  {$I-} Rewrite(PgmFile) {$I+};
  IF (IOresult<>0) THEN
    BEGIN
      Msg := '**ERROR**  Cannot open file ' + Filename;
      ShowMsgOpt(Msg, LightRed, 0);
      SaveMsg(Msg);
      Stop_Flag := true;
    END
  ELSE {It's Cool: Go ahead}
    IF Debug THEN WriteLn;

 {-------( Write out #defines for IN/OUT/Variables    )--------}

    {----( Write the PgmDesc definition )----}
    PgmLine := '#define ' + 'PgmDesc ' + PgmDescription;
    WriteLn(PgmFile,PgmLine);

    {----( Write the SampleTime definition )----}
    Str(SampleTime, UtilStr);
    PgmLine := '#define ' + 'SampleTime ' + UtilStr;
    WriteLn(PgmFile,PgmLine);

    {----( Write the QuietMODE definition )----}
    IF QuietMODE THEN UtilStr := 'TRUE' ELSE UtilStr := 'FALSE';
    PgmLine := '#define ' + 'QuietMODE ' + UtilStr;
    WriteLn(PgmFile,PgmLine);

    {----( Write the LastEdit definition )----}
    Str(LastEdit, UtilStr);
    PgmLine := '#define ' + 'LastEdit ' + UtilStr;
    WriteLn(PgmFile,PgmLine);

    {----( Write the OUTPUT Names definition )----}
    FOR n := 0 to (MaxIOBits - 1) DO
      BEGIN
        WITH OutputNamesArray[n] DO
          BEGIN
            Str(n, UtilStr);
            PgmLine := '#define ' + 'OUT' + UtilStr + ' ' + IOName;
            IF Inverted THEN PgmLine := PgmLine + ' I';
          END;
        WriteLn(PgmFile,PgmLine);
      END;

    {----( Write the INPUT Names definition )----}
    FOR n := 0 to (MaxIOBits - 1) DO
      BEGIN
        WITH InputNamesArray[n] DO
          BEGIN
            Str(n, UtilStr);
            PgmLine := '#define ' + 'IN' + UtilStr + ' ' + IOName;
            IF Inverted THEN PgmLine := PgmLine + ' I';
          END;
        WriteLn(PgmFile,PgmLine);
      END;

 {-------( Write out the Program lines from StepArray )--------}
    NumLines := PgmTopStep;
    FOR n := 1 to NumLines DO
      BEGIN {Processing File Lines}

      {-------( Read out of Array )---}
      WITH StepArray[n] DO
        BEGIN

      {-------( Build Output Line  )---}
          PgmLine    := StepName    +
                        NextStep    +
                        StepString  +
                        StepOper    ;

          WriteLn(PgmFile,PgmLine);
        END;
      END;{Processing Save File Lines}
  Close(PgmFile);
END;{SavePgm}


{------------------------------------------------------------------------}
{-----( EditIOEntry: Edit an IOEntry on the display)---------------------}
{------------------------------------------------------------------------}

PROCEDURE EditIOEntry;

VAR
  ValStr     : String[4];
  WorkStr    : String[32];
  N          : word;
  Done       : Boolean;
  Ch         : Char;

BEGIN
   Done := FALSE;
   UserString := '';
   IF EditOut THEN Window(2, YCur, 9, YCur)
              ELSE Window(71, YCur, 78, YCur);

    {--( EDIT IOName )--}
      GoToXY(1,1);
      ResponseType := ReadlnPF(UserString, KeyCode, FALSE );{Get user input}
      IF (ResponseType = Esc) THEN Done := TRUE;
      IF NOT Done AND (ResponseType <> Tab) THEN
        BEGIN
          IF (ResponseType = CommandStr) THEN
            UserString := Upper(UserString);
            IF (Length(UserString) > 0) THEN   {No action if just 'enter'}
              BEGIN
                WHILE (Length(UserString) < 8)
                  DO UserString := UserString + ' ';
                IF EditOut THEN OutPutNamesArray[Bit2Edit].IOName := UserString
                           ELSE InputNamesArray [Bit2Edit].IOName := UserString;
                UserString := '';
              END;
          ShowIONames;
        END; {Not Done}
END;


(*Ŀ
   MAIN PROGRAM:                                                           
  *)
{M1}
BEGIN

  {---------( INITIALIZE )-----------------------------------------}
  BMState          := STOPPED;
  TestLooping      := FALSE;
  ForceNextStep    := FALSE;
  QuietMODE        := FALSE;
  Samples          := 0;
  SampleTime       := 100;                 {Milliseconds}
  Step             := 1;

  MsgWritePtr      := 0;    {Saved messages}
  MsgReadPtr       := 1;
  MsgMaxPtr        := 1;

  OutputByteInit   := OUTPUTBYTESTART;
  OutputByte       := OutputByteInit;
  OutputByteInvert := $00;
  InputByteInvert  := $00;

  InitArrays;

  InitScreen;

  UpDateScreen := TRUE;

  {---------( TRY A CALL )-----------------------------------------}
(*Call('beep.exe', '', Rtn);*)
(*Call('pcutil.com', 'TS', Rtn);*)
(*  Call('command.com','try.bat',Rtn);*)

  REPEAT {-----------------( MAIN LOOP IN PROGRAM )-----------------------}

   {---( EXECUTE THE CURRENT STEP )----}

   Delay(SampleTime);
   INC(Samples);

   ShowState;
   IF ((BMState = RUNNING) OR (BMState = STEPPING))
     THEN
       BEGIN
         IF NOT QuietMODE THEN Click;
         ExecuteStep;
       END;

   ExecuteIO;

   {---( DISPLAY CHANGED DATA )----}
   GOTOXY(17,16);
   Show8Bits(OutputByte);
   GOTOXY(51,16);
   Show8Bits(InputByte);

   ShowSamples;
   ShowStepNum;

   {----( DISPLAY TIMERS IF IN USE )-----}
   StrobeHRTimers;
   ConvertHRTimers;
   FOR N := 1 TO MAXTIMERS DO
     BEGIN
       IF TimerInUseArray[n] THEN ShowTimer(n);
     END;

  {---------( GET USER INPUT, CHECK, HANDLE DIFFERENT TYPES )------}
   UserString := '';

   ResponseType := ReadlnPF(UserString, KeyCode, TRUE );  {Get user input}

   IF (ResponseType <> NoKey) THEN
     BEGIN
       IF (ResponseType = CommandStr) THEN  (*??*)
         BEGIN
           Statement := UserString;
           UserString := '';
           GoToXY(2,17);
           Write( Statement);
           GoToXY(2,17);
         END;
     {------------( Look for ESC Key )-------------------------}

(*     IF (ResponseType = Esc) THEN Done := TRUE;*)

     {------------( Look for Extended Keys )-------------------------}

       IF (ResponseType = ExtendedKey) THEN
         BEGIN
           GoToXY(61,3);
           IF (KeyCode = UPARROW) THEN
             BEGIN
               write('Up Arrow');
               DEC(Step2Show);
               IF (Step2Show <= 0) THEN Step2Show := 1;
               UpdateScreen := TRUE;
             END;
           IF (KeyCode = DOWNARROW) THEN
             BEGIN
               write('Dn Arrow');
               INC(Step2Show);
               IF (Step2Show > (MAXSTEP - 5)) THEN Step2Show := (MAXSTEP-5);
               UpdateScreen := TRUE;
             END;

           IF (KeyCode = PGUP     ) THEN
             BEGIN
               write('PageUp');
               DEC(Step2Show, 5);
               IF (Step2Show <= 1) THEN Step2Show := 1;
               UpdateScreen := TRUE;
             END;

           IF (KeyCode = PGDN     ) THEN
             BEGIN
               write('PageDn');
               INC(Step2Show, 5);
               IF (Step2Show > (MAXSTEP-5)) THEN Step2Show := (MAXSTEP-5);
               UpdateScreen := TRUE;
             END;

           IF (KeyCode = CTRL_PGUP     ) THEN
             BEGIN
               write('Msg?+');
               INC(MsgReadPtr);
               IF (MsgReadPtr > MsgMaxPtr) THEN MsgReadPtr := 1;
               BlankMsg;
               ShowMsg(MsgSaveArray[MsgReadPtr]);
             END;


           IF (KeyCode = CTRL_PGDN     ) THEN
             BEGIN
               write('Msg?-');
               DEC(MsgReadPtr);
               IF (MsgReadPtr < 1) THEN MsgReadPtr := MsgMaxPtr;
               BlankMsg;
               ShowMsg(MsgSaveArray[MsgReadPtr]);
             END;


           IF (KeyCode = INS      ) THEN
             BEGIN
               write('Insert');
               InsertStepEntry(Step2Show + 2);
               UpDateScreen := TRUE;
             END;
           IF (KeyCode = DEL      ) THEN
             BEGIN
               write('Delete');
               ShowMsgOpt('DELETE THIS LINE ??(Y/N)', LightRed, BLINKIT);
               KbChar := ReadKey;
               BlankMsg;
               IF (KbChar = 'Y') OR (KbChar = 'y') THEN
                 DeleteStepEntry(Step2Show + 2);
               UpDateScreen := TRUE;
             END;
           IF (KeyCode = HOME     ) THEN
             BEGIN
               write('Home');
               Step2Show := 1;
               UpdateScreen := TRUE;
             END;
           IF (KeyCode = _END     ) THEN
             BEGIN
               write('End');
          (*     Step2Show := MAXSTEP-5;*)
                 Step2Show := PgmTopStep -2;
               UpdateScreen := TRUE;
             END;
          {--( TOGGLE OUTPUT BITS MANUALLY )----}
           IF (KeyCode = ALT_1    ) THEN ToggleBit(7);
           IF (KeyCode = ALT_2    ) THEN ToggleBit(6);
           IF (KeyCode = ALT_3    ) THEN ToggleBit(5);
           IF (KeyCode = ALT_4    ) THEN ToggleBit(4);
           IF (KeyCode = ALT_5    ) THEN ToggleBit(3);
           IF (KeyCode = ALT_6    ) THEN ToggleBit(2);
           IF (KeyCode = ALT_7    ) THEN ToggleBit(1);
           IF (KeyCode = ALT_8    ) THEN ToggleBit(0);


           IF (KeyCode = ALT_D    ) THEN
             BEGIN
               BlankMsg;
               PgmDescription := GetUserString('PgmDesc= ','',17,24);
               ShowMsg('PgmDesc: ' + PgmDescription);
             END;
           IF (KeyCode = ALT_C    ) THEN
             BEGIN
               write('Copy');
               CopyStepEntry(Step2Show + 2);
               ShowMsg('Line Saved: <Alt>P to PASTE');
             END;
           IF (KeyCode = ALT_P    ) THEN
             BEGIN
               write('Paste');
               PasteStepEntry(Step2Show + 2);
               UpdateScreen := TRUE;
             END;
           IF (KeyCode = ALT_N    ) THEN write('Names');
           IF (KeyCode = ALT_R    ) THEN
             BEGIN
               write('RESTART');
               Step := 2;
               Step2Show := 1;
               FOR n := 1 TO MAXCOUNTERS DO CounterArray[n] := 0;
               ClearCounterArea;
               UpDateScreen := TRUE;
             END;
           IF (KeyCode = ALT_A    ) THEN
             BEGIN
               BlankMsg;
               UtilStr := GetUserString('SampleTime = ','',17,24);
               Val(UtilStr, SampleTime, RtnCode);
               Msg := 'SampleTime: ' + UtilStr + ' Milliseconds';
               ShowMsg(Msg);
               SaveMsg(Msg);
             END;
           IF (KeyCode = ALT_S    ) THEN write('STOP');
           IF (KeyCode = ALT_T    ) THEN write('STEP');
           IF (KeyCode = ALT_B    ) THEN write('Breakpoint');
           IF (KeyCode = ALT_H    ) THEN write('Help');
           IF (KeyCode = ALT_Q    )
           THEN
             BEGIN
               write('Quiet');
               IF QuietMode THEN QuietMODE := FALSE ELSE QuietMODE := TRUE;
             END;

           GoToXY(2,17);
         END;{Extended Keys}

     {------------( Look for Extended Keys )-------------------------}

       IF (ResponseType = PFKEY) THEN
         BEGIN
           GoToXY(61,2);
           IF (KeyCode = F1       ) THEN write('F1:Help');
           IF (KeyCode = F2       ) THEN
             BEGIN
               write('F2:File');
               BlankMsg;
               PgmFileName := GetUserString('Filename = ', '',
                                            17,24);
               IF (POS('.', PgmFileName) = 0)
                 THEN PgmFileName := PgmFileName + '.bim';
               InitArrays;
               LoadPgm(PgmFileName);
               OldPgmFileName := PgmFileName;
               ShowIONames;
             END;

           IF (KeyCode = F3       ) THEN
             BEGIN
               ShowMsgOpt('F3:Quit ??(Y/N)', LightRed, BLINKIT);
               KbChar := ReadKey;
               BlankMsg;
               IF (KbChar = 'Y') OR (KbChar = 'y') THEN Done := TRUE;
             END;

           IF (KeyCode = F4       ) THEN
             BEGIN
               write('F4:EDIT');
               EditStepEntry;
             END;

           IF (KeyCode = ALT_F4       ) THEN
             BEGIN
               write('EDIT I/O');
               EditIOEntry;
             END;


           IF (KeyCode = F5       ) THEN
             BEGIN
               write('F5:RUN ');
               IF (BMState = ENDED) THEN Step := 2; {??}
               BMState := RUNNING;
             END;

           IF (KeyCode = F6       ) THEN
             BEGIN
               write('F6:STOP');
               BMState := STOPPED;
             END;

           IF (KeyCode = F7       ) THEN
             BEGIN
               write('F7:STEP');
               IF (BMState = ENDED) THEN Step := 2; {??}
               BMState := STEPPING;
             END;

           IF (KeyCode = CTRL_F7       ) THEN
             BEGIN
               write('FORCE:STEP');
               ForceNextStep := TRUE;
             END;

           IF (KeyCode = F8       ) THEN
             BEGIN
               write('F8:BRKP');
               BMState := BREAKING;
             END;

           IF (KeyCode = F9       ) THEN
             BEGIN
               write('F9:    ');
             END;

           IF (KeyCode = F10      ) THEN
             BEGIN
               write('F10:Save');
               BlankMsg;
               PgmFileName := GetUserString('Filename = ', '',
                                            17,24);
               IF (PgmFileName[1] = '=') THEN PgmFileName := OldPgmFileName;
               IF (POS('.', PgmFileName) = 0)
                 THEN PgmFileName := PgmFileName + '.bim';
               SavePgm(PgmFileName);
               Msg := 'SAVED AS: ' + PgmFileName;
               ShowMsg(Msg);
               SaveMsg(Msg);
             END;

           GoToXY(61,2);
         END;{PFKeys}
       END;{Not NoKey}

 {-------( UPDATE PROGRAM DISPLAY SCREEN )-------- }
   IF UpDateScreen THEN
    BEGIN
       Window(PWIN_S_ULX, PWIN_S_ULY, PWIN_LRX, PWIN_LRY);
       FOR n := Step2Show TO (Step2Show + 5) DO
       BEGIN
         IF (n = Step2Show + 2)
           THEN Appearance := APPEAR_HIGH
           ELSE Appearance := APPEAR_NORM;
         ShowStepEntry(n , Appearance); Writeln;
       END;
       Window(1,1,80,25);
       UpDateScreen := FALSE;
    END;


  UNTIL Done;
{----------------( END OF MAIN PROGRAM LOOP )--------------------}

(*  Readln(KbChar);*)
  Textmode(Co80);                        {Put the Display back to normal }


END.
